"""
This module exposes dependencies used by lightkube-models

These dependencies are here because we may decide to replace dataclasses with something else in the future
"""

__all__ = ["DictMixin", "dataclass", "field"]
from dataclasses import dataclass, field

from .dataclasses_dict import DataclassDictMixIn as DictMixin



import asyncio.subprocess
import json
import os
import ssl
import subprocess
from dataclasses import asdict, dataclass, field
from typing import AsyncGenerator, Callable, Dict, Generator, List, Mapping, Optional, Sequence, Tuple, overload

import httpx

from ..core.exceptions import ConfigError
from .kubeconfig import SingleConfig, StrOrPath
from .models import Cluster, FileStr, User, UserExec


@dataclass
class ConnectionParams:
    """All connection parameters used by Client and AsyncClient"""

    timeout: Optional[httpx.Timeout] = field(default_factory=lambda: httpx.Timeout(10))
    trust_env: bool = True
    transport: Optional[httpx.BaseTransport] = None
    proxy: Optional[str] = None
    http2: bool = False

    def httpx_params(self, config: SingleConfig) -> dict:
        base_url = config.cluster.server
        assert config.user, "Missing user"
        verify = verify_cluster(config.cluster, config.user, config.abs_file, trust_env=self.trust_env)
        auth = user_auth(config.user)
        return dict(base_url=base_url, verify=verify, auth=auth, **asdict(self))


def Client(config: SingleConfig, conn_parameters: ConnectionParams) -> httpx.Client:
    return httpx.Client(**conn_parameters.httpx_params(config))


def AsyncClient(config: SingleConfig, conn_parameters: ConnectionParams) -> httpx.AsyncClient:
    return httpx.AsyncClient(**conn_parameters.httpx_params(config))


class BearerAuth(httpx.Auth):
    def __init__(self, token: str) -> None:
        self._bearer = f"Bearer {token}"

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        request.headers["Authorization"] = self._bearer
        yield request


async def async_check_output(command: Sequence[str], env: Mapping[str, str]) -> bytes:
    PIPE = asyncio.subprocess.PIPE
    proc = await asyncio.create_subprocess_exec(*command, env=env, stdin=None, stdout=PIPE, stderr=PIPE)
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        raise ConfigError(f"Exec {command[0]} returned {proc.returncode}: {stderr.decode()}")
    return stdout


def sync_check_output(command: Sequence[str], env: Mapping[str, str]) -> bytes:
    proc = subprocess.Popen(command, env=env, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        raise ConfigError(f"Exec {command[0]} returned {proc.returncode}: {stderr.decode()}")
    return stdout


class ExecAuth(httpx.Auth):
    # The param name `exec` shadows the Python builtin, but changing the kwarg name would potentially break users code
    def __init__(self, exec: UserExec) -> None:  # noqa: A002
        self._exec = exec
        self._last_bearer: Optional[str] = None

    def _prepare(self) -> Tuple[List[str], Dict[str, str]]:
        _exec = self._exec
        if _exec.apiVersion not in (
            "client.authentication.k8s.io/v1alpha1",
            "client.authentication.k8s.io/v1beta1",
        ):
            raise ConfigError(f"auth exec api version {_exec.apiVersion} not implemented")
        cmd_env_vars = dict(os.environ)
        if _exec.env:
            cmd_env_vars.update((var.name, var.value) for var in _exec.env)
        # TODO: add support for passing KUBERNETES_EXEC_INFO env var
        # https://github.com/kubernetes/community/blob/master/contributors/design-proposals/auth/kubectl-exec-plugins.md
        args = _exec.args if _exec.args else []
        return [_exec.command, *args], cmd_env_vars

    def sync_auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        if self._last_bearer:
            request.headers["Authorization"] = self._last_bearer
            response = yield request
            if response.status_code != 401:
                return

        command, env = self._prepare()
        output = sync_check_output(command, env=env)
        token = json.loads(output)["status"]["token"]
        request.headers["Authorization"] = self._last_bearer = f"Bearer {token}"
        yield request

    async def async_auth_flow(self, request: httpx.Request) -> AsyncGenerator[httpx.Request, httpx.Response]:
        if self._last_bearer:
            request.headers["Authorization"] = self._last_bearer
            response = yield request
            if response.status_code != 401:
                return

        command, env = self._prepare()
        output = await async_check_output(command, env=env)
        token = json.loads(output)["status"]["token"]
        request.headers["Authorization"] = self._last_bearer = f"Bearer {token}"
        yield request


@overload
def user_auth(user: User) -> Optional[httpx.Auth]: ...


@overload
def user_auth(user: None) -> None: ...


def user_auth(user: Optional[User]) -> Optional[httpx.Auth]:
    if user is None:
        return None

    if user.token is not None:
        return BearerAuth(user.token)

    if user.exec:
        return ExecAuth(user.exec)

    if user.username and user.password:
        return httpx.BasicAuth(user.username, user.password)

    if user.auth_provider:
        raise ConfigError("auth-provider not supported")

    # TODO: Is this intended? Previously, a return statement was just missing entirely.
    return None


def user_cert(user: User, abs_file: Callable[[StrOrPath], StrOrPath]) -> Optional[Tuple[StrOrPath, StrOrPath]]:
    """Extract user certificates"""
    if user.client_cert or user.client_cert_data:
        return (
            FileStr(user.client_cert_data) or abs_file(user.client_cert),  # type: ignore[arg-type]
            FileStr(user.client_key_data) or abs_file(user.client_key),  # type: ignore[arg-type]
        )
    return None


def verify_cluster(
    cluster: Cluster,
    user: User,
    abs_file: Callable[[StrOrPath], StrOrPath],
    trust_env: bool = True,
) -> ssl.SSLContext:
    """setup certificate verification"""
    if cluster.certificate_auth:
        ctx = ssl.create_default_context(cafile=abs_file(cluster.certificate_auth))
    elif cluster.certificate_auth_data:
        ctx = ssl.create_default_context(cafile=FileStr(cluster.certificate_auth_data))
    else:
        ctx = httpx.create_ssl_context(verify=not cluster.insecure, trust_env=trust_env)
    cert = user_cert(user, abs_file)
    if cert:
        ctx.load_cert_chain(*cert)
    return ctx



import base64
import tempfile
from dataclasses import dataclass, field
from typing import IO, Dict, List, Optional, overload

from ..core.dataclasses_dict import DataclassDictMixIn


class FileStr(str):
    handler: Optional[IO[bytes]] = None

    # TODO: Remove non-typechecking compliant tricks here (a `__new__` function that doesn't return the class itself,
    #   but either `str` (which at least is a subclass) or `None`) and use factory methods instead of overwriting
    #   `__new__` or just have a custom class with a `__init__` method.
    @overload
    def __new__(cls, data: None) -> None: ...  # type: ignore[misc]
    @overload
    def __new__(cls, data: str) -> str: ...  # type: ignore[misc]

    def __new__(cls, data: Optional[str]) -> Optional[str]:  # type: ignore[misc]
        if data is None:
            return None

        f = tempfile.NamedTemporaryFile()
        f.write(base64.b64decode(data))
        f.flush()
        file = str.__new__(cls, f.name)
        file.handler = f
        return file

    def __del__(self) -> None:
        if self.handler:
            self.handler.close()
            self.handler = None


@dataclass
class Context(DataclassDictMixIn):
    cluster: str
    user: Optional[str] = None
    namespace: Optional[str] = None


@dataclass
class NameValue(DataclassDictMixIn):
    name: str
    value: str


@dataclass
class UserExec(DataclassDictMixIn):
    apiVersion: str
    command: str
    env: List[NameValue] = field(default_factory=list)
    args: List[str] = field(default_factory=list)
    installHint: Optional[str] = None


@dataclass
class User(DataclassDictMixIn):
    exec: Optional[UserExec] = None
    username: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None
    auth_provider: Optional[Dict] = field(metadata={"json": "auth-provider"}, default=None)
    client_cert: Optional[str] = field(metadata={"json": "client-certificate"}, default=None)
    client_cert_data: Optional[str] = field(metadata={"json": "client-certificate-data"}, default=None)
    client_key: Optional[str] = field(metadata={"json": "client-key"}, default=None)
    client_key_data: Optional[str] = field(metadata={"json": "client-key-data"}, default=None)


@dataclass
class Cluster(DataclassDictMixIn):
    """
    Attributes:
      server: the server name
    """

    server: str = "http://localhost:8080"
    certificate_auth: Optional[str] = field(metadata={"json": "certificate-authority"}, default=None)
    certificate_auth_data: Optional[str] = field(metadata={"json": "certificate-authority-data"}, default=None)
    insecure: bool = field(metadata={"json": "insecure-skip-tls-verify"}, default=False)



from typing import Any, Dict, Iterable, List, Tuple, Union

from lightkube import operators

FIELDS_SUPPORT = ("equal", "not_equal", "not_in")
FIELDS_SUPPORT_STR = ", ".join(f'"{fs}"' for fs in FIELDS_SUPPORT)

_VALUE_TYPE = Union[None, str, operators.Operator[Any], Iterable[str]]


def build_selector(pairs: Union[List[Tuple[str, _VALUE_TYPE]], Dict[str, _VALUE_TYPE]], for_fields: bool = False) -> str:
    res = []
    pairs_it: Iterable[Tuple[str, _VALUE_TYPE]]
    if not isinstance(pairs, list):
        pairs_it = pairs.items()
    else:
        pairs_it = pairs
    for k, v in pairs_it:
        if v is None:
            v = operators.exists()
        elif isinstance(v, str):
            v = operators.equal(v)
        elif isinstance(v, Iterable):
            v = operators.in_(v)

        if not isinstance(v, operators.Operator):
            raise ValueError(f"selector value '{v}' should be str, None, Iterable or instance of operator")

        if for_fields and v.op_name not in FIELDS_SUPPORT:
            raise ValueError(f"parameter 'fields' only support operators {FIELDS_SUPPORT_STR}")

        if for_fields and v.op_name == "not_in":  # not_in can be implement using several !=
            assert isinstance(v, operators.SequenceOperator)
            for item in v.value:
                res.append(operators.not_equal(item).encode(k))
        else:
            res.append(v.encode(k))
    return ",".join(res)



import enum
import typing
from dataclasses import dataclass


class PatchType(enum.Enum):
    """
    Attributes:
      JSON: Execute a json patch
      MERGE: Execute a json merge patch
      STRATEGIC: Execute a strategic merge patch
      APPLY: Execute a [server side apply](https://kubernetes.io/docs/reference/using-api/server-side-apply/)
    """

    JSON = "application/json-patch+json"
    MERGE = "application/merge-patch+json"
    STRATEGIC = "application/strategic-merge-patch+json"
    APPLY = "application/apply-patch+yaml"


class CascadeType(enum.Enum):
    """
    Attributes:
      ORPHAN: orphan the dependents
      BACKGROUND: allow the garbage collector to delete the dependents in the background
      FOREGROUND: a cascading policy that deletes all dependents in the foreground
    """

    ORPHAN = "Orphan"
    BACKGROUND = "Background"
    FOREGROUND = "Foreground"


class OnErrorAction(enum.Enum):
    """
    Attributes:
      RETRY: Retry to perform the API call again from the last version
      STOP: Stop silently the iterator
      RAISE: Raise the error on the caller scope
    """

    RETRY = 0
    STOP = 1
    RAISE = 2


@dataclass
class OnErrorResult:
    action: OnErrorAction
    sleep: float = 0


OnErrorHandler = typing.Callable[[Exception, int], OnErrorResult]


def on_error_raise(e: Exception, count: int) -> OnErrorResult:
    """Raise the error on the caller scope"""
    return OnErrorResult(OnErrorAction.RAISE)


def on_error_stop(e: Exception, count: int) -> OnErrorResult:
    """Stop silently the iterator"""
    return OnErrorResult(OnErrorAction.STOP)


def on_error_retry(e: Exception, count: int) -> OnErrorResult:
    """Retry to perform the API call again from the last version"""
    return OnErrorResult(OnErrorAction.RETRY)



from .core.exceptions import ApiError, ConditionError, ConfigError, LoadResourceError, NotReadyError, ObjectDeleted

__all__ = [
    "ApiError",
    "ConditionError",
    "ConfigError",
    "LoadResourceError",
    "NotReadyError",
    "ObjectDeleted",
]



from typing import Generic, Iterable, TypeVar

__all__ = ["equal", "exists", "in_", "not_equal", "not_exists", "not_in"]

T_Y = TypeVar("T_Y")


class Operator(Generic[T_Y]):
    def __init__(self, op_name: str, op: str, value: T_Y) -> None:
        self.op = op
        self.value: T_Y = value
        self.op_name = op_name

    def encode(self, key: str) -> str:
        return f"{key}{self.op}{self.value}"


class SequenceOperator(Operator[Iterable[str]]):
    def encode(self, key: str) -> str:
        return f"{key} {self.op} ({','.join(self.value)})"


class BinaryOperator(Operator[str]):
    pass


class UnaryOperator(Operator[None]):
    def __init__(self, op_name: str, op: str) -> None:
        super().__init__(op_name, op, value=None)

    def encode(self, key: str) -> str:
        return f"{self.op}{key}"


def in_(values: Iterable[str]) -> SequenceOperator:
    return SequenceOperator("in_", "in", sorted(values))


def not_in(values: Iterable[str]) -> SequenceOperator:
    return SequenceOperator("not_in", "notin", sorted(values))


def exists() -> UnaryOperator:
    return UnaryOperator("exists", "")


def not_exists() -> UnaryOperator:
    return UnaryOperator("not_exists", "!")


def equal(value: str) -> BinaryOperator:
    return BinaryOperator("equal", "=", value)


def not_equal(value: str) -> BinaryOperator:
    return BinaryOperator("not_equal", "!=", value)



import asyncio
import dataclasses
import json
import time
from dataclasses import dataclass
from typing import Any, AsyncIterable, AsyncIterator, Dict, Iterable, Iterator, Optional, Tuple, Type, TypeVar, Union

import httpx

from ..config import client_adapter
from ..config.kubeconfig import DEFAULT_KUBECONFIG, KubeConfig, SingleConfig
from ..types import OnErrorAction, OnErrorHandler, PatchType, on_error_raise
from . import resource as r
from .exceptions import ApiError, NotReadyError

ALL_NS = "*"


def transform_exception(e: httpx.HTTPError):
    if isinstance(e, httpx.HTTPStatusError) and e.response.headers["Content-Type"] == "application/json":
        return ApiError(request=e.request, response=e.response)
    return e


METHOD_MAPPING = {
    "delete": "DELETE",
    "deletecollection": "DELETE",
    "get": "GET",
    "global_list": "GET",
    "global_watch": "GET",
    "list": "GET",
    "patch": "PATCH",
    "post": "POST",
    "put": "PUT",
    "watch": "GET",
}


@dataclass
class BasicRequest:
    method: str
    url: str
    response_type: Any
    params: Dict[str, str] = dataclasses.field(default_factory=dict)
    data: Any = None
    headers: Dict[str, str] = None


class WatchDriver:
    def __init__(self, br: BasicRequest, build_request, lazy):
        self._version = br.params.get("resourceVersion")
        self._convert = br.response_type.from_dict
        self._br = br
        self._build_request = build_request
        self._lazy = lazy

    def get_request(self, timeout):
        br = self._br
        if self._version is not None:
            br.params["resourceVersion"] = self._version
        return self._build_request(br.method, br.url, params=br.params, timeout=timeout)

    def process_one_line(self, line):
        line = json.loads(line)
        tp = line["type"]
        obj = line["object"]
        if tp == "ERROR":
            raise ApiError(status=obj)
        self._version = obj["metadata"]["resourceVersion"]
        return tp, self._convert(obj, lazy=self._lazy)


T = TypeVar("T")


class ListIterable(Iterable[T]):
    _resourceVersion: Optional[str] = None

    @property
    def resourceVersion(self) -> str:
        """Returns the resource version at which the collection was constructed.
        Note: this property is only available after the iteration started and will raise NotReadyError otherwise
        """
        if self._resourceVersion:
            return self._resourceVersion
        raise NotReadyError("resourceVersion", "only available after the iteration started")

    def __init__(self, inner_iter: Iterator[Tuple[str, Iterator[T]]]) -> None:
        self._inner_iter = inner_iter

    def __iter__(self) -> Iterator[T]:
        for rv, chunk in self._inner_iter:
            self._resourceVersion = rv
            yield from chunk


class ListAsyncIterable(AsyncIterable[T]):
    _resourceVersion: Optional[str] = None

    @property
    def resourceVersion(self) -> str:
        """Returns the resource version at which the collection was constructed.
        Note: this property is only available after the iteration started and will raise NotReadyError otherwise
        """
        if self._resourceVersion:
            return self._resourceVersion
        raise NotReadyError("resourceVersion", "only available after the iteration started")

    def __init__(self, inner_iter: AsyncIterator[Tuple[str, Iterator[T]]]) -> None:
        self._inner_iter = inner_iter

    async def __aiter__(self) -> AsyncIterator[T]:
        async for rv, chunk in self._inner_iter:
            self._resourceVersion = rv
            for item in chunk:
                yield item


class GenericClient:
    AdapterClient = staticmethod(client_adapter.Client)

    def __init__(
        self,
        conn_params: client_adapter.ConnectionParams,
        config: Union[SingleConfig, KubeConfig] = None,
        namespace: Optional[str] = None,
        lazy=True,
        field_manager: Optional[str] = None,
        dry_run: bool = False,
    ):
        self._watch_timeout = httpx.Timeout(conn_params.timeout)
        self._watch_timeout.read = None
        self._lazy = lazy
        if config is None and conn_params.trust_env:
            config = KubeConfig.from_env().get()
        elif config is None and not conn_params.trust_env:
            config = KubeConfig.from_file(DEFAULT_KUBECONFIG).get()
        elif isinstance(config, KubeConfig):
            config = config.get()

        self.config = config
        self._client = self.AdapterClient(config, conn_params)
        self._field_manager = field_manager
        self._dry_run = dry_run
        self.namespace = namespace if namespace else config.namespace

    def prepare_request(
        self,
        method,
        res: Optional[Type[r.Resource]] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> BasicRequest:
        if params is not None:
            params = {k: v for k, v in params.items() if v is not None}
        else:
            params = {}
        if headers is not None:
            headers = {k: v for k, v in headers.items() if v is not None}
        data = None
        if res is None:
            if obj is None:
                raise ValueError("At least a resource or an instance of a resource need to be provided")
            res = obj.__class__

        namespaced = issubclass(res, (r.NamespacedResource, r.NamespacedSubResource))

        if namespace == ALL_NS:
            if not issubclass(res, r.NamespacedResourceG):
                raise ValueError(f"Class {res} doesn't support global {method}")
            if method not in ("list", "watch"):
                raise ValueError("Only methods 'list' and 'watch' can be called for all namespaces")
            real_method = "global_watch" if watch else "global_" + method
        else:
            real_method = "watch" if watch else method

        api_info = r.api_info(res)
        if real_method not in api_info.verbs:
            if watch:
                raise ValueError(f"Resource '{res.__name__}' is not watchable")
            else:
                raise ValueError(f"method '{method}' not supported by resource '{res.__name__}'")

        if watch:
            params["watch"] = "true"

        if api_info.parent is None:
            base = api_info.resource
        else:
            base = api_info.parent

        if base.group == "":
            path = ["api", base.version]
        else:
            path = ["apis", base.group, base.version]

        if namespaced and namespace != ALL_NS:
            if method in ("post", "put") and obj.metadata.namespace is not None:
                if namespace is None:
                    namespace = obj.metadata.namespace
                elif namespace != obj.metadata.namespace:
                    raise ValueError(
                        f"The namespace value '{namespace}' differ from the "
                        f"namespace in the object metadata '{obj.metadata.namespace}'"
                    )
            if namespace is None:
                namespace = self.namespace
            path.extend(["namespaces", namespace])

        if method in ("post", "put", "patch"):
            if self._field_manager is not None and "fieldManager" not in params:
                params["fieldManager"] = self._field_manager
            if self._dry_run is True and "dryRun" not in params:
                params["dryRun"] = "All"
            if method == "patch" and headers["Content-Type"] == PatchType.APPLY.value and "fieldManager" not in params:
                raise ValueError('Parameter "field_manager" is required for PatchType.APPLY')
            if obj is None:
                raise ValueError("obj is required for post, put or patch")

            if method == "patch" and not isinstance(obj, r.Resource):
                data = obj
            else:
                data = obj.to_dict()
                # The following block, ensures that apiVersion and kind are always set.
                # this is needed as k8s fails if this data are not provided for objects derived by CRDs (Issue #27)
                if "apiVersion" not in data:
                    data["apiVersion"] = api_info.resource.api_version
                if "kind" not in data:
                    data["kind"] = api_info.resource.kind

        path.append(api_info.plural)
        if method in ("delete", "get", "patch", "put") or api_info.action:
            if name is None and method == "put":
                name = obj.metadata.name
            if name is None:
                raise ValueError("resource name not defined")
            path.append(name)

        if api_info.action:
            path.append(api_info.action)

        http_method = METHOD_MAPPING[method]
        if http_method == "DELETE":
            res = None

        return BasicRequest(
            method=http_method,
            url="/".join(path),
            params=params,
            response_type=res,
            data=data,
            headers=headers,
        )

    @staticmethod
    def raise_for_status(resp):
        try:
            resp.raise_for_status()
        except httpx.HTTPError as e:
            raise transform_exception(e) from e

    def build_adapter_request(self, br: BasicRequest):
        return self._client.build_request(br.method, br.url, params=br.params, json=br.data, headers=br.headers)

    def convert_to_resource(self, res: Type[r.Resource], item: dict) -> r.Resource:
        resource_def = r.api_info(res).resource
        item.setdefault("apiVersion", resource_def.api_version)
        item.setdefault("kind", resource_def.kind)
        return res.from_dict(item, lazy=self._lazy)

    def handle_response(self, method, resp, br):
        self.raise_for_status(resp)
        res = br.response_type
        if res is None:
            # TODO: delete/deletecollection actions normally return a Status object, we may want to return it as well
            return
        data = resp.json()
        if method == "list":
            if "metadata" in data and data["metadata"].get("continue"):
                cont = True
                br.params["continue"] = data["metadata"]["continue"]
            else:
                cont = False
            try:
                rv = data["metadata"]["resourceVersion"]
            except KeyError:
                rv = None
            return (
                cont,
                rv,
                (self.convert_to_resource(res, obj) for obj in data["items"]),
            )
        else:
            if res is not None:
                return self.convert_to_resource(res, data)


class GenericSyncClient(GenericClient):
    def send(self, req, stream=False):
        return self._client.send(req, stream=stream)

    def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                for line in resp.iter_lines():
                    yield wd.process_one_line(line)
            except httpx.HTTPError as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    time.sleep(handle_error.sleep)
                continue

    def request(
        self,
        method,
        res: Optional[Type[r.Resource]] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> Any:
        br = self.prepare_request(method, res, obj, name, namespace, watch, headers=headers, params=params)
        req = self.build_adapter_request(br)
        resp = self.send(req)
        return self.handle_response(method, resp, br)

    def list_chunks(self, br: BasicRequest) -> Iterator[Tuple[str, Iterator]]:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = self.send(req)
            cont, rv, chunk = self.handle_response("list", resp, br)
            yield rv, chunk

    def list(self, br: BasicRequest) -> ListIterable:
        return ListIterable(self.list_chunks(br))


class GenericAsyncClient(GenericClient):
    AdapterClient = staticmethod(client_adapter.AsyncClient)

    async def send(self, req, stream=False):
        return await self._client.send(req, stream=stream)

    async def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = await self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                async for line in resp.aiter_lines():
                    yield wd.process_one_line(line)
            except httpx.HTTPError as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    await asyncio.sleep(handle_error.sleep)
                continue
            finally:
                await resp.aclose()

    async def request(
        self,
        method,
        res: Optional[Type[r.Resource]] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> Any:
        br = self.prepare_request(method, res, obj, name, namespace, watch, headers=headers, params=params)
        req = self.build_adapter_request(br)
        resp = await self.send(req)
        return self.handle_response(method, resp, br)

    async def list_chunks(self, br: BasicRequest) -> AsyncIterator[Tuple[str, Iterator]]:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = await self.send(req)
            cont, rv, chunk = self.handle_response("list", resp, br)
            yield rv, chunk

    def list(self, br: BasicRequest) -> ListAsyncIterable:
        return ListAsyncIterable(self.list_chunks(br))

    async def close(self):
        await self._client.aclose()
